from typing import Tuple, List, Dict, Optional
from pathlib import Path
import json

import pytest
import pandas as pd

import pynblast

from .constants import DATA_DIR


@pytest.fixture
def smat_path() -> Path:
    return DATA_DIR / "smat_fcwb.csv"


@pytest.fixture
def score_mat_tup(smat_path) -> pynblast.ScoreMatrix:
    return pynblast.ScoreMatrix.read(smat_path)


def parse_points(fpath: Path) -> pd.DataFrame:
    df = pd.read_csv(fpath, sep=",", usecols=[1, 2, 3], header=0)
    df.columns = [c.lower() for c in df.columns]
    return df


@pytest.fixture
def points() -> List[Tuple[str, pd.DataFrame]]:
    p_dir = DATA_DIR / "points"
    return sorted((fpath.stem, parse_points(fpath)) for fpath in p_dir.glob("*.csv"))


@pytest.fixture
def arena(score_mat_tup):
    return pynblast.NblastArena(*score_mat_tup, k=5)


@pytest.fixture
def arena_names(arena, points) -> Tuple[pynblast.NblastArena, Dict[int, str]]:
    """Tuple of (arena, {idx: "name"})"""
    idx_to_name = dict()
    for name, df in points:
        idx = arena.add_points(df.to_numpy())
        idx_to_name[idx] = name
    return arena, idx_to_name


@pytest.fixture
def expected_nblast() -> pd.DataFrame:
    """
    Query on row index, target on column index.
    """
    p_dir = DATA_DIR / "points"
    names = sorted(p.stem for p in p_dir.glob("*.csv"))
    df = pd.read_csv(DATA_DIR / "kcscores.csv", index_col=0)
    df = df[names]
    df = df.loc[names]
    return df.T


@pytest.fixture
def dotprops() -> List[Tuple[str, pd.DataFrame]]:
    p_dir = DATA_DIR / "dotprops"
    return sorted((p.stem, pd.read_csv(p, index_col=0)) for p in p_dir.glob("*.csv"))


@pytest.fixture
def skeleton() -> List[Tuple[int, Optional[int], float, float, float]]:
    with open(DATA_DIR / "arbors" / "example.json") as f:
        return [tuple(row) for row in json.load(f)["data"]]


@pytest.fixture
def resampler(skeleton) -> pynblast.ResamplingArbor:
    return pynblast.ResamplingArbor(skeleton)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `nblast` package."""
from typing import Tuple, Dict

import pytest
import pandas as pd
import numpy as np

from pynblast import NblastArena, Idx, ScoreMatrix, rectify_tangents, Symmetry

EPSILON = 0.001


def test_read_smat(smat_path):
    dist_bins, dot_bins, arr = ScoreMatrix.read(smat_path)
    assert arr.shape == (len(dist_bins), len(dot_bins))
    df = pd.read_csv(smat_path, index_col=0, header=0)
    assert np.allclose(arr, df.to_numpy())


def test_construction(score_mat_tup):
    NblastArena(*score_mat_tup)


def test_insertion(arena: NblastArena, points):
    df = points[0][1]
    arena.add_points(df.to_numpy())


def test_points_conserved(arena: NblastArena, points):
    p1 = points[0][1].to_numpy()
    idx = arena.add_points(p1)
    p2 = arena.points(idx)
    assert np.allclose(p1, p2)


@pytest.mark.parametrize(
    ["test", "expected"],
    [
        ([1, 1, 1], [1, 1, 1]),
        ([-1, -1, -1], [1, 1, 1]),
        ([-1, 1, 1], [1, -1, -1]),
        ([0, 1, 1], [0, 1, 1]),
        ([0, -1, 1], [0, 1, -1]),
        ([0, 0, 1], [0, 0, 1]),
        ([0, 0, -1], [0, 0, 1]),
        ([0, 0, 0], [0, 0, 0]),
    ],
)
def test_rectify_tangents(test, expected):
    out = rectify_tangents(np.array([test]))
    expected = np.array([expected])
    assert np.allclose(out, expected)


def test_tangents(arena, dotprops):
    df = dotprops[0][1]
    points = df[["points." + d for d in "XYZ"]].to_numpy()
    expected_tangents = rectify_tangents(
        df[[f"vect.{n}" for n in range(1, 4)]].to_numpy()
    )
    idx = arena.add_points(points)
    tangents = arena.tangents(idx, True)
    assert np.allclose(tangents, expected_tangents)


def test_query(arena_names: Tuple[NblastArena, Dict[int, str]]):
    arena, _ = arena_names

    result = arena.query_target(Idx(0), Idx(1))
    assert result


def all_v_all(arena, names, normalize=False, symmetry=None, threads=None):
    q = list(names)
    return arena.queries_targets(q, q, normalize, symmetry=symmetry, threads=threads)


def test_all_v_all(arena_names: Tuple[NblastArena, Dict[int, str]]):
    out = all_v_all(*arena_names, threads=None)
    assert len(out) == len(arena_names[1]) ** 2


def test_all_v_all_par(arena_names: Tuple[NblastArena, Dict[int, str]]):
    ser = all_v_all(*arena_names, threads=None)
    par = all_v_all(*arena_names, threads=0)
    assert ser == par


def test_normed(arena_names):
    out = all_v_all(*arena_names, normalize=True)
    for (q, t), v in out.items():
        if q == t:
            assert v == 1


def test_self_hit(arena, points):
    """Check that the self-hit results are correct.

    Ensures that explicit self-hits (with speed optimisations) return
    practically the same result as implicit self-hits (testing two
    identical neurons against each other).

    As a side-effect, proves that tangent vectors are unit-length, as
    this is assumed by the self-hit optimisations.
    """
    _name, df = points[0]
    idx1 = arena.add_points(df.to_numpy())
    idx2 = arena.add_points(df.to_numpy())
    true_self = arena.query_target(idx1, idx1)
    different_self = arena.query_target(idx1, idx2)
    assert different_self == pytest.approx(true_self, 0.0001)


def test_self_hit_calc(arena_names):
    arena, _ = arena_names
    explicit = arena.self_hit(0)
    implicit = arena.query_target(0, 0)
    assert explicit == pytest.approx(implicit, abs=EPSILON)


def test_normed_calc(arena_names):
    arena, _ = arena_names
    self_hit = arena.query_target(0, 0)
    non_norm = arena.query_target(0, 1)
    norm = arena.query_target(0, 1, normalize=True)

    assert non_norm / self_hit == pytest.approx(norm)


def test_symmetric(arena_names):
    out = all_v_all(*arena_names, symmetry="arithmetic_mean")
    for (q, t), v in out.items():
        assert out[(t, q)] == v


@pytest.mark.parametrize("sym", list(Symmetry))
def test_symmetry_enum(arena_names, sym):
    arena, _ = arena_names
    out1 = arena.query_target(0, 1, symmetry=sym)
    out2 = arena.query_target(1, 0, symmetry=sym)
    assert out1 == out2


def test_prepop_tangents(points, arena):
    p = points[0][1].to_numpy()
    idx0 = arena.add_points(p)
    tangents = arena.tangents(idx0)
    idx1 = arena.add_points_tangents(p, tangents)
    assert arena.query_target(idx0, idx1, normalize=True) == pytest.approx(
        1.0, abs=EPSILON
    )



from typing import List, Tuple, Dict, Iterator, Optional

import numpy as np

from .pynblast import ArenaWrapper
from .util import Idx, raise_if_none, rectify_tangents, Symmetry


class NblastArena:
    """
    Class for creating and keeping track of many neurons for comparison with NBLAST.
    """

    threads = 0
    DEFAULT_K = 20

    def __init__(
        self,
        dist_bins: List[float],
        dot_bins: List[float],
        score_mat: np.ndarray,
        threads: Optional[int] = -1,
        k=None,
    ):
        """
        The required arguments describe a lookup table which is used to convert
        ``(distance, abs_dot_product)`` tuples into a score for a single
        point match.
        The ``*_bins`` arguments describe the upper bounds of the bins
        (although anything greater than the uppermost bound is considered to belong
        to the top bin).
        ``score_mat`` is the table of values, in dist-major order.

        For example, if the lookup table was stored as a pandas dataframe,
        where the distance bins were in the left margin and the absolute dot
        product bins in the top margin, the object would be instantiated by

        >>> arena = NblastArena(df.index, df.columns, df.to_numpy())

        See the ``ScoreMatrix`` namedtuple for convenience.

        ``k`` gives the number of points to use when calculating tangents.

        ``threads`` sets the default number of threads to use.
        If `threads` < 0 (default), use the class' ``threads`` attribute (default 0).
        If it is 0, use the number of threads available.
        If it is > 0, use at most that many.
        If it is ``None``, run in serial.
        """
        self.threads = self._parse_threads(threads)
        self.k = k or self.DEFAULT_K

        if score_mat.shape != (len(dist_bins), len(dot_bins)):
            raise ValueError("Bin thresholds do not match score matrix")
        score_vec = score_mat.flatten().tolist()
        self._impl = ArenaWrapper(dist_bins, dot_bins, score_vec, self.k)

    def add_points(self, points: np.ndarray) -> Idx:
        """Add an Nx3 point cloud to the arena representing a neuron.
        Tangents will be calculated automatically.

        Returns an integer index which is used to refer to that neuron later,
        for queries etc..
        """
        if points.ndim != 2 or points.shape[1] != 3:
            raise ValueError("Points must have shape Nx3")
        return self._impl.add_points(points.tolist())

    def add_points_tangents(self, points: np.ndarray, tangents: np.ndarray) -> Idx:
        """Add an Nx3 point cloud representing a neuron, with pre-calculated tangents.
        Tangents are assumed to be unit-length and in the same order as the points.

        Returns an integer index which is used to refer to that neuron later.
        """
        if points.shape != tangents.shape or points.ndim != 2 or points.shape[1] != 3:
            raise ValueError("Points and tangents must have the same shape, Nx3")

        return self._impl.add_points_tangents(points.tolist(), tangents.tolist())

    def _parse_threads(self, threads: Optional[int]) -> Optional[int]:
        if threads is None:
            return None
        elif threads < 0:
            return self.threads
        else:
            return int(threads)

    def query_target(
        self,
        query_idx: Idx,
        target_idx: Idx,
        normalize: bool = False,
        symmetry: Optional[Symmetry] = None,
    ) -> float:
        """Query one neuron against another,
        using the indices generated when they were added.

        If ``normalize`` is true, divide the raw score by the query neuron's
        self-hit score.

        If ``symmetry`` is not ``None``, additionally calculate the score from target
        to query (normalizing if requested), and apply some function to make the
        operation commutative/ symmetric
        (see the ``Symmetry`` enum for available functions).
        """
        out = self._impl.query_target(query_idx, target_idx, bool(normalize), symmetry)
        return raise_if_none(out, query_idx, target_idx)

    def queries_targets(
        self,
        query_idxs: List[Idx],
        target_idxs: List[Idx],
        normalize: bool = False,
        symmetry: Optional[Symmetry] = None,
        threads: Optional[int] = -1,
    ) -> Dict[Tuple[Idx, Idx], float]:
        """Query all combinations of some query neurons against some target neurons.

        See the ``query_target`` method for more details on
        ``normalize`` and ``symmetry``.
        See the ``__init__`` method for more details on ``threads``.
        """
        threads = self._parse_threads(threads)
        return self._impl.queries_targets(
            query_idxs, target_idxs, bool(normalize), symmetry, threads
        )

    def all_v_all(
        self, normalize=False, symmetry=None, threads: Optional[int] = -1,
    ) -> Dict[Tuple[Idx, Idx], float]:
        """Query all loaded neurons against each other.

        See the ``query_target`` method for more details on
        ``normalize`` and ``symmetry``.
        See the ``__init__`` method for more details on ``threads``.
        """
        threads = self._parse_threads(threads)
        return self._impl.all_v_all(bool(normalize), symmetry, threads)

    def __len__(self) -> int:
        return self._impl.len()

    def __iter__(self) -> Iterator[Idx]:
        for idx in range(len(self)):
            yield Idx(idx)

    def self_hit(self, idx) -> float:
        """Get the raw score for querying the given neuron against itself.

        N.B. this is much faster that ``arena.query_target(n, n)``.
        """
        out = self._impl.self_hit(idx)
        return raise_if_none(out, idx)

    def points(self, idx) -> np.ndarray:
        """Return a copy of the points associated with the indexed neuron.

        Point order is arbitrary, but consistent with the order returned by the
        ``.tangents`` method.
        """
        return np.array(raise_if_none(self._impl.points(idx), idx))

    def tangents(self, idx, rectify=False) -> np.ndarray:
        """Return a copy of the tangents associated with the indexed neuron.

        Tangent order is arbitrary, but consistent with the order returned by the
        ``.points`` method.
        """
        out = np.array(raise_if_none(self._impl.tangents(idx), idx))
        return rectify_tangents(out, True) if rectify else out



from typing import NewType
import enum

import numpy as np


class StrEnum(str, enum.Enum):
    def __new__(cls, *args):
        for arg in args:
            if not isinstance(arg, (str, enum.auto)):
                raise TypeError(
                    "Values of StrEnums must be strings: {} is a {}".format(
                        repr(arg), type(arg)
                    )
                )
        return super().__new__(cls, *args)

    def __str__(self):
        return self.value

    # The first argument to this function is documented to be the name of the
    # enum member, not `self`:
    # https://docs.python.org/3.6/library/enum.html#using-automatic-values
    def _generate_next_value_(name, *_):
        return name


Idx = NewType("Idx", int)


def raise_if_none(result, *idxs):
    if result is None:
        raise IndexError(f"Index(es) not in arena: {idxs}")
    return result


class Symmetry(StrEnum):
    """Enum of strategies for making an NBLAST query symmetric.

    i.e. the name of a function to apply to the forward and reverse scores
    together to generate a comparable score.
    """

    ARITHMETIC_MEAN = "arithmetic_mean"
    GEOMETRIC_MEAN = "geometric_mean"
    HARMONIC_MEAN = "harmonic_mean"
    MIN = "min"
    MAX = "max"


def rectify_tangents(orig: np.ndarray, inplace=False) -> np.ndarray:
    """Normalises orientation of tangents.

    Makes the first nonzero element positive.
    """
    if not inplace:
        orig = orig.copy()

    prev_zero = np.full(len(orig), True, dtype=bool)
    for this_col in orig.T:
        to_flip = np.logical_and(prev_zero, this_col < 0)
        orig[to_flip, :] *= -1
        prev_zero = np.logical_and(prev_zero, this_col == 0)
        if not prev_zero.any():
            break

    return orig



# -*- coding: utf-8 -*-

"""Top-level package for nblast-rs."""

__author__ = """Chris L. Barnes"""
__email__ = "chrislloydbarnes@gmail.com"

from .pynblast import get_version as _get_version, ResamplingArbor

__version__ = _get_version()
__version_info__ = tuple(int(n) for n in __version__.split("."))

from .util import rectify_tangents, Idx, Symmetry
from .arena import NblastArena
from .score_matrix import ScoreMatrix

__all__ = [
    "NblastArena",
    "ScoreMatrix",
    "Symmetry",
    "rectify_tangents",
    "Idx",
    "ResamplingArbor",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `nblast` package."""
import pytest
import pandas as pd
import numpy as np

EPSILON = 0.001


def dict_to_df(idxs_to_score, idx_to_name):
    names = sorted(idx_to_name.values())
    size = len(idx_to_name)
    df = pd.DataFrame(np.full((size, size), np.nan), index=names, columns=names)
    for (q_idx, t_idx), score in idxs_to_score.items():
        q_name = idx_to_name[q_idx]
        t_name = idx_to_name[t_idx]
        df.loc[q_name][t_name] = score

    assert not np.isnan(df.to_numpy()).any()
    return df


def test_vs_r(arena_names, expected_nblast):
    arena, idx_to_name = arena_names
    idxs_to_score = arena.all_v_all()

    for (q_idx, t_idx), score in idxs_to_score.items():
        q_name = idx_to_name[q_idx]
        t_name = idx_to_name[t_idx]
        score = idxs_to_score[(q_idx, t_idx)]
        expected_score = expected_nblast.loc[q_name][t_name]
        assert score == pytest.approx(expected_score, abs=EPSILON)



"""Type stubs for classes to be used directly from pynblast"""
from typing import List, Optional, Tuple

class ResamplingArbor:
    def __init__(self, table: List[Tuple[int, Optional[int], float, float, float]]): ...
    def prune_at(self, ids: List[int]) -> int: ...
    def prune_branches_containing(self, ids: List[int]) -> int: ...
    def prune_below_strahler(self, threshold: int) -> int: ...
    def prune_beyond_branches(self, threshold: int) -> int: ...
    def prune_beyond_steps(self, threshold: int) -> int: ...
    def points(self) -> List[Tuple[float, float, float]]: ...
    def skeleton(self) -> List[Tuple[int, Optional[int], float, float, float]]: ...



from typing import List, NamedTuple
import csv

import numpy as np


def parse_interval(s):
    no_brackets = s.strip("([]) ")
    if not no_brackets:
        return None
    low_high = no_brackets.split(",")
    if len(low_high) == 1:
        return float(low_high)
    else:
        return float(low_high[-1])


class ScoreMatrix(NamedTuple):
    dist_thresholds: List[float]
    dot_thresholds: List[float]
    values: np.ndarray

    def to_df(self):
        import pandas as pd

        return pd.DataFrame(self.values, self.dist_thresholds, self.dot_thresholds)

    @classmethod
    def read(cls, fpath, **csv_kwargs):
        """Read a precomputed score matrix from a CSV.

        Assumes the matrix has distance bins on the row index,
        and dot product bins on the column index.
        Assumes the col/row indices are strings describing
        the upper bound of the bin,
        or the interval that bin covers, e.g. "(0.75,1.5]"
        (N.B. all brackets are ignored; upper value is assumed to be closed
        and lower value is assumed to be 0 or the upper value of the previous bin).

        Returns a tuple of distance thresholds, dot product thresholds,
        and the array of values.
        """
        with open(fpath) as f:
            reader = csv.reader(f, **csv_kwargs)
            dot_bins = [parse_interval(s) for s in next(reader)[1:]]

            dist_bins = []
            data = []
            for idx, row in enumerate(reader, 1):
                dist_bins.append(parse_interval(row[0]))
                row_data = [float(s) for s in row[1:]]
                if len(row_data) != len(dot_bins):
                    raise ValueError(
                        f"Line {idx} has {len(row_data)} values; "
                        f"expected {len(dot_bins)}"
                    )
                data.append(row_data)

        return cls(dist_bins, dot_bins, np.array(data))



from contextlib import contextmanager

import pytest

from pynblast import ResamplingArbor


def get_nodes(resampler):
    return {tup[0] for tup in resampler.skeleton()}


def assert_nodes(resampler, contains=None, not_contains=None):
    if not contains:
        contains = set()
    if not not_contains:
        not_contains = set()

    nodes = get_nodes(resampler)

    assert nodes.issuperset(contains)
    assert nodes.isdisjoint(not_contains)


def test_construction(skeleton):
    ResamplingArbor(skeleton)


def test_prune_at(resampler):
    prune_at = [23329984]
    with assert_removes_nodes(resampler, not_contains=prune_at):
        resampler.prune_at(prune_at)
    assert_nodes(resampler, not_contains=prune_at)


@pytest.mark.skip("Failure in underlying library")
def test_prune_branches_containing(resampler):
    prune_containing = [6208611, 6208792, 3537598]
    with assert_removes_nodes(resampler, not_contains=prune_containing):
        resampler.prune_branches_containing(prune_containing)


@contextmanager
def assert_removes_nodes(resampler, contains=None, not_contains=None):
    before = get_nodes(resampler)
    yield
    after = get_nodes(resampler)
    assert after.issubset(before)
    assert len(after) < len(before)
    if contains:
        assert after.issuperset(contains)
    if not_contains:
        assert after.isdisjoint(not_contains)


def test_prune_below_strahler(resampler):
    with assert_removes_nodes(resampler):
        resampler.prune_below_strahler(2)


def test_prune_beyond_branches(resampler):
    with assert_removes_nodes(resampler):
        resampler.prune_beyond_branches(3)


def test_points(resampler):
    resampler.points()


def test_points_resample(resampler):
    before = resampler.points()
    after = resampler.points(1000)
    assert len(after) < len(before)


def test_skeleton(resampler):
    resampler.skeleton()
